# -*- coding: utf-8 -*-

from zope.interface import Interface


class IEventLineGetter(Interface):

    def got_event_line(self, line, timestamp):
        """Process line from event log.

        :param line: line which describes event.
        :type line: str.

        :param timestamp: timestamp provided by log watching service.
        :type timestamp: datetime.
        """


class IPilotService(Interface):

    def user_join(self, info):
        """
        """

    def user_left(self, info):
        """
        """

    def seat_occupied(self, info):
        """
        """

    def weapons_loaded(self, info):
        """
        """

    def was_killed(self, info):
        """
        """

    def was_shot_down(self, info):
        """
        """

    def selected_army(self, info):
        """
        """

    def went_to_menu(self, info):
        """
        """


class IObjectsService(Interface):

    def was_destroyed(self, info):
        """
        """



# -*- coding: utf-8 -*-

from twisted.trial.unittest import TestCase

from il2ds_middleware.constants import MISSION_STATUS, PILOT_LEAVE_REASON
from il2ds_middleware.parser import (ConsoleParser, EventLogParser,
    DeviceLinkParser, )
from il2ds_middleware.tests.service import PilotService, ObjectsService


class ConsoleParserTestCase(TestCase):

    def setUp(self):
        self.pilot_srvc = PilotService()
        self.parser = ConsoleParser(self.pilot_srvc)
        self.pilot_srvc.startService()

    def tearDown(self):
        return self.pilot_srvc.stopService()

    def test_server_info(self):
        datas = [
            "Type: Local server",
            "Name: Test server",
            "Description: Horus test server", ]
        result = self.parser.server_info(datas)
        self.assertIsInstance(result, dict)
        self.assertEqual(result.get('type'), "Local server")
        self.assertEqual(result.get('name'), "Test server")
        self.assertEqual(result.get('description'), "Horus test server")

    def test_mission_status(self):
        status, mission = self.parser.mission_status([
            "Mission NOT loaded", ])
        self.assertEqual(status, MISSION_STATUS.NOT_LOADED)
        self.assertEqual(mission, None)

        status, mission = self.parser.mission_status([
            "Mission: net/dogfight/test.mis is Loaded", ])
        self.assertEqual(status, MISSION_STATUS.LOADED)
        self.assertEqual(mission, "net/dogfight/test.mis")

        status, mission = self.parser.mission_status([
            "Mission: net/dogfight/test.mis is Playing", ])
        self.assertEqual(status, MISSION_STATUS.PLAYING)
        self.assertEqual(mission, "net/dogfight/test.mis")

        datas = ["some fake data", ]
        result = self.parser.mission_status(datas)
        self.assertIsInstance(result, list)
        self.assertEqual(len(datas), len(result))
        self.assertEqual(datas[0], result[0])

    def test_mission_load(self):
        datas = [
            "Loading mission net/dogfight/test.mis...",
            "Load bridges",
            "Load static objects",
            "##### House without collision (3do/Tree/Tree2.sim)",
            "##### House without collision "
                "(3do/Buildings/Port/Floor/live.sim)",
            "##### House without collision "
                "(3do/Buildings/Port/BaseSegment/live.sim)",
            "Mission: net/dogfight/test.mis is Loaded", ]
        status, mission = self.parser.mission_load(datas)
        self.assertEqual(status, MISSION_STATUS.LOADED)
        self.assertEqual(mission, "net/dogfight/test.mis")

    def test_mission_begin(self):
        status, mission = self.parser.mission_begin([
            "Mission: net/dogfight/test.mis is Playing", ])
        self.assertEqual(status, MISSION_STATUS.PLAYING)
        self.assertEqual(mission, "net/dogfight/test.mis")

    def test_mission_end(self):
        status, mission = self.parser.mission_end([
            "Mission: net/dogfight/test.mis is Loaded", ])
        self.assertEqual(status, MISSION_STATUS.LOADED)
        self.assertEqual(mission, "net/dogfight/test.mis")

    def test_mission_destroy(self):
        status, mission = self.parser.mission_destroy([
            "Mission NOT loaded", ])
        self.assertEqual(status, MISSION_STATUS.NOT_LOADED)
        self.assertEqual(mission, None)

    def test_user_joined(self):
        self.parser.parse_line(
            "socket channel '0' start creating: ip 192.168.1.2:21000")
        self.parser.parse_line("Chat: --- user0 joins the game.")
        self.parser.parse_line(
            "socket channel '0', ip 192.168.1.2:21000, user0, "
            "is complete created.")
        self.assertEqual(len(self.pilot_srvc.joined), 1)
        info = self.pilot_srvc.joined[0]

        self.assertIsInstance(info, dict)
        self.assertEqual(info.get('channel'), 0)
        self.assertEqual(info.get('ip'), "192.168.1.2")
        self.assertEqual(info.get('callsign'), "user0")

    def test_user_left(self):
        self.parser.parse_line(
            "socketConnection with 192.168.1.2:21000 on channel 0 lost.  "
            "Reason: ")
        self.parser.parse_line("Chat: --- user0 has left the game.")
        self.assertEqual(len(self.pilot_srvc.left), 1)
        info = self.pilot_srvc.left[0]

        self.assertIsInstance(info, dict)
        self.assertEqual(info.get('channel'), 0)
        self.assertEqual(info.get('ip'), "192.168.1.2")
        self.assertEqual(info.get('callsign'), "user0")
        self.assertEqual(info.get('reason'), PILOT_LEAVE_REASON.DISCONNECTED)

        self.parser.parse_line(
            "socketConnection with 192.168.1.3:21000 on channel 1 lost.  "
            "Reason: You have been kicked from the server.")
        self.parser.parse_line("Chat: --- user1 has left the game.")
        self.assertEqual(len(self.pilot_srvc.left), 2)
        info = self.pilot_srvc.left[1]

        self.assertIsInstance(info, dict)
        self.assertEqual(info.get('channel'), 1)
        self.assertEqual(info.get('ip'), "192.168.1.3")
        self.assertEqual(info.get('callsign'), "user1")
        self.assertEqual(info.get('reason'), PILOT_LEAVE_REASON.KICKED)


class EventLogParserTestCase(TestCase):

    def setUp(self):
        self.pilot_srvc = PilotService()
        self.obj_srvc = ObjectsService()
        self.parser = EventLogParser((self.pilot_srvc, self.obj_srvc))
        self.pilot_srvc.startService()

    def tearDown(self):
        self.pilot_srvc.stopService()
        self.obj_srvc.stopService()

    def test_occupied_seat(self):
        data = "user0:A6M2-21(0) seat occupied by user0 at 100.99 200.99"
        self.parser.parse_line(data)

        self.assertEqual(len(self.pilot_srvc.occupied), 1)
        result = self.pilot_srvc.occupied[0]

        self.assertIsInstance(result, dict)
        self.assertEqual(result.get('callsign'), "user0")
        self.assertEqual(result.get('seat'), 0)
        self.assertEqual(result.get('aircraft'), "A6M2-21")
        self.assertIsInstance(result.get('pos'), dict)
        self.assertEqual(result['pos'].get('x'), 100.99)
        self.assertEqual(result['pos'].get('y'), 200.99)

    def test_weapons_loaded(self):
        data = "user0:A6M2-21 loaded weapons '1xdt' fuel 100%"
        self.parser.parse_line(data)

        self.assertEqual(len(self.pilot_srvc.weapons), 1)
        result = self.pilot_srvc.weapons[0]

        self.assertIsInstance(result, dict)
        self.assertEqual(result.get('callsign'), "user0")
        self.assertEqual(result.get('aircraft'), "A6M2-21")
        self.assertEqual(result.get('weapons'), "1xdt")
        self.assertEqual(result.get('fuel'), 100)

    def test_was_killed(self):
        data = "user0:A6M2-21(0) was killed at 100.99 200.99"
        self.parser.parse_line(data)

        self.assertEqual(len(self.pilot_srvc.killed), 1)
        result = self.pilot_srvc.killed[0]

        self.assertIsInstance(result, dict)
        self.assertEqual(result.get('callsign'), "user0")
        self.assertEqual(result.get('seat'), 0)
        self.assertEqual(result.get('aircraft'), "A6M2-21")
        self.assertIsInstance(result.get('pos'), dict)
        self.assertEqual(result['pos'].get('x'), 100.99)
        self.assertEqual(result['pos'].get('y'), 200.99)

    def test_was_shot_down(self):
        data = "user0:A6M2-21 shot down by user1:B5N2 at 100.99 200.99"
        self.parser.parse_line(data)

        self.assertEqual(len(self.pilot_srvc.shot_down), 1)
        result = self.pilot_srvc.shot_down[0]

        self.assertIsInstance(result, dict)
        self.assertIsInstance(result.get('victim'), dict)
        self.assertEqual(result['victim'].get('callsign'), "user0")
        self.assertEqual(result['victim'].get('aircraft'), "A6M2-21")
        self.assertIsInstance(result.get('attacker'), dict)
        self.assertEqual(result['attacker'].get('callsign'), "user1")
        self.assertEqual(result['attacker'].get('aircraft'), "B5N2")
        self.assertIsInstance(result.get('pos'), dict)
        self.assertEqual(result['pos'].get('x'), 100.99)
        self.assertEqual(result['pos'].get('y'), 200.99)

    def test_selected_army(self):
        data = "user0 selected army Red at 100.99 200.99"
        self.parser.parse_line(data)

        self.assertEqual(len(self.pilot_srvc.army_select), 1)
        result = self.pilot_srvc.army_select[0]

        self.assertIsInstance(result, dict)
        self.assertEqual(result.get('callsign'), "user0")
        self.assertEqual(result.get('army'), 'Red')
        self.assertIsInstance(result.get('pos'), dict)
        self.assertEqual(result['pos'].get('x'), 100.99)
        self.assertEqual(result['pos'].get('y'), 200.99)

    def test_went_to_menu(self):
        data = "user0 entered refly menu"
        self.parser.parse_line(data)

        self.assertEqual(len(self.pilot_srvc.to_menu), 1)
        result = self.pilot_srvc.to_menu[0]

        self.assertIsInstance(result, dict)
        self.assertEqual(result.get('callsign'), "user0")

    def test_was_destroyed(self):
        data = "0_Static destroyed by landscape at 100.99 200.99"
        self.parser.parse_line(data)

        self.assertEqual(len(self.obj_srvc.destroyed), 1)
        result = self.obj_srvc.destroyed[0]

        self.assertIsInstance(result, dict)
        self.assertEqual(result.get('victim'), "0_Static")
        self.assertEqual(result.get('attacker'), "landscape")
        self.assertIsInstance(result.get('pos'), dict)
        self.assertEqual(result['pos'].get('x'), 100.99)
        self.assertEqual(result['pos'].get('y'), 200.99)


class DeviceLinkParserTestCase(TestCase):

    def setUp(self):
        self.parser = DeviceLinkParser()

    def tearDown(self):
        self.parser = None

    def test_pilot_count(self):
        result = self.parser.pilot_count('0')
        self.assertEqual(result, 0)

    def test_pilot_pos(self):
        result = self.parser.pilot_pos('0:user0;100;200;300')
        self.assertIsInstance(result, dict)
        self.assertEqual(result.get('idx'), 0)
        self.assertEqual(result.get('callsign'), "user0")
        self.assertIsInstance(result.get('pos'), dict)
        self.assertEqual(result['pos'].get('x'), 100)
        self.assertEqual(result['pos'].get('y'), 200)
        self.assertEqual(result['pos'].get('z'), 300)

    def test_all_pilots_pos(self):
        datas = ["{0}:user{0};{1};{2};{3}".format(
            i, i*100, i*200, i*300) for i in xrange(10)]
        results = self.parser.all_pilots_pos(datas)
        self.assertIsInstance(results, list)
        self.assertEqual(len(datas), len(results))
        for i in xrange(len(results)):
            result = results[i]
            self.assertIsInstance(result, dict)
            self.assertEqual(result.get('idx'), i)
            self.assertEqual(result.get('callsign'), "user{0}".format(i))
            self.assertIsInstance(result.get('pos'), dict)
            self.assertEqual(result['pos'].get('x'), i*100)
            self.assertEqual(result['pos'].get('y'), i*200)
            self.assertEqual(result['pos'].get('z'), i*300)

    def test_static_count(self):
        result = self.parser.static_count('0')
        self.assertEqual(result, 0)

    def test_static_pos(self):
        result = self.parser.static_pos('0:0_Static;100;200;300')
        self.assertIsInstance(result, dict)
        self.assertEqual(result.get('idx'), 0)
        self.assertEqual(result.get('name'), "0_Static")
        self.assertIsInstance(result.get('pos'), dict)
        self.assertEqual(result['pos'].get('x'), 100)
        self.assertEqual(result['pos'].get('y'), 200)
        self.assertEqual(result['pos'].get('z'), 300)

    def test_all_static_pos(self):
        datas = ["{0}:{0}_Static;{1};{2};{3}".format(
            i, i*100, i*200, i*300) for i in xrange(10)]
        results = self.parser.all_static_pos(datas)
        self.assertIsInstance(results, list)
        self.assertEqual(len(datas), len(results))
        for i in xrange(len(results)):
            result = results[i]
            self.assertIsInstance(result, dict)
            self.assertEqual(result.get('idx'), i)
            self.assertEqual(result.get('name'), "{0}_Static".format(i))
            self.assertIsInstance(result.get('pos'), dict)
            self.assertEqual(result['pos'].get('x'), i*100)
            self.assertEqual(result['pos'].get('y'), i*200)
            self.assertEqual(result['pos'].get('z'), i*300)



# -*- coding: utf-8 -*-

RX_POS = " at (\d+.\d+) (\d+.\d+)"

RX_SEAT_OCCUPIED = r"(.+):(.+)\((\d+)\) seat occupied by .+" + RX_POS
RX_WEAPONS_LOADED = r"(.+):(.+) loaded weapons \'(.+)\' fuel (\d+)%"
RX_KILLED = r"(.+):(.+)\((.+)\) was killed" + RX_POS
RX_SHOT_DOWN = r"(.+):(.+) shot down by (.+):(.+)" + RX_POS
RX_SELECTED_ARMY = r"(.+) selected army (.+)" + RX_POS
RX_WENT_TO_MENU = r"(.+) entered refly menu"
RX_DESTROYED = r"(.+) destroyed by (.+)" + RX_POS
RX_USER_JOIN = r"socket channel '(\d+)', ip (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}):\d+, (.+), is complete created."
RX_USER_LEFT = r"socketConnection with (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}):\d+ on channel (\d+) lost.  Reason: (.*)"



# -*- coding: utf-8 -*-

import datetime

from twisted.application.internet import TimerService
from twisted.application.service import Service
from twisted.internet import defer

from zope.interface import implementer

from il2ds_middleware.interface.service import IPilotService, IObjectsService


class LogWatchingBaseService(TimerService):

    receiver = None

    def __init__(self, log_path, interval=1):
        self.log_file = None
        self.log_path = log_path
        TimerService.__init__(self, interval, self.do_watch)

    def do_watch(self):
        self.log_file.seek(self.log_file.tell())
        for line in self.log_file.readlines():
            self.got_line(line)

    def got_line(self, line):
        raise NotImplementedError

    def startService(self):
        if self.log_file is not None:
            return
        self.log_file = open(self.log_path, 'r')
        TimerService.startService(self)

    def stopService(self):
        if self.log_file is None:
            return defer.succeed(None)
        else:
            self.log_file.close()
            self.log_file = None
            return TimerService.stopService(self)


class LogWatchingService(LogWatchingBaseService):

    def got_line(self, line):
        if self.receiver is not None:
            timestamp = datetime.datetime.now()
            self.receiver.got_event_line(line, timestamp)


class ClientBaseService(Service):

    def __init__(self, console=None):
        self.console = console


@implementer(IPilotService)
class PilotBaseService(ClientBaseService):

    def user_join(self, info):
        raise NotImplementedError

    def user_left(self, info):
        raise NotImplementedError

    def seat_occupied(self, info):
        raise NotImplementedError

    def weapons_loaded(self, info):
        raise NotImplementedError

    def was_killed(self, info):
        raise NotImplementedError

    def was_shot_down(self, info):
        raise NotImplementedError

    def selected_army(self, info):
        raise NotImplementedError

    def went_to_menu(self, info):
        raise NotImplementedError


@implementer(IObjectsService)
class ObjectsBaseService(ClientBaseService):

    def was_destroyed(self, info):
        raise NotImplementedError



# -*- coding: utf-8 -*-

REQ_SERVER_INFO = "server"

REQ_MISSION_STATUS = "mission"
REQ_MISSION_LOAD = "mission LOAD {0}"
REQ_MISSION_BEGIN = "mission BEGIN"
REQ_MISSION_END = "mission END"
REQ_MISSION_DESTROY = "mission DESTROY"



# -*- coding: utf-8 -*-

from twisted.internet.defer import Deferred
from twisted.internet.protocol import ServerFactory
from twisted.protocols.basic import LineReceiver
from twisted.python import log

from il2ds_middleware.protocol import DeviceLinkProtocol


class ConsoleServer(LineReceiver):

    service = None

    def connectionMade(self):
        self.factory.client_connected(self)

    def connectionLost(self, reason):
        self.factory.client_left(self)

    def lineReceived(self, line):
        if self.service is not None:
            self.service.parse_line(line)

    def message(self, message):
        self.sendLine(message + '\\n')


class ConsoleServerFactory(ServerFactory):

    protocol = ConsoleServer

    def __init__(self):
        self.on_connected = Deferred()
        self.on_connection_lost = Deferred()

    def client_connected(self, client):
        if self.on_connected is not None:
            d, self.on_connected = self.on_connected, None
            d.callback(client)

    def client_left(self, client):
        if self.on_connection_lost is not None:
            d, self.on_connection_lost = self.on_connection_lost, None
            d.callback(client)


class DeviceLinkServerProtocol(DeviceLinkProtocol):

    receiver = None

    def requests_received(self, requests, address):
        if self.receiver is not None:
            self.receiver(requests, address, self)



# -*- coding: utf-8 -*-

from twisted.internet import defer
from twisted.internet.error import ConnectionDone
from twisted.internet.protocol import (ClientFactory, DatagramProtocol, )
from twisted.protocols.basic import LineOnlyReceiver
from twisted.python import log

from il2ds_middleware.constants import (DEVICE_LINK_OPCODE as DL_OPCODE,
    DEVICE_LINK_PREFIXES, DEVICE_LINK_CMD_SEPARATOR as DL_CMD_SEP,
    DEVICE_LINK_ARGS_SEPARATOR as DL_ARGS_SEP, DEVICE_LINK_CMD_GROUP_MAX_SIZE,
    REQUEST_TIMEOUT, REQUEST_MISSION_LOAD_TIMEOUT, )
from il2ds_middleware.requests import (
    REQ_SERVER_INFO, REQ_MISSION_STATUS, REQ_MISSION_LOAD, REQ_MISSION_BEGIN,
    REQ_MISSION_END, REQ_MISSION_DESTROY, )


class ConsoleClient(LineOnlyReceiver):

    def __init__(self, parser=None):
        self._parser = parser
        self._request_id = 0
        self._responce_id = None
        # { request_id: ([results], deferred, timeout, ), }
        self._requests = {}

    def connectionMade(self):
        self.factory.clientConnectionMade(self)

    def _generate_request_id(self):
        self._request_id += 1
        return self._request_id

    def _make_request(self, d, timeout_value):
        rid = self._generate_request_id()

        def on_timeout(_):
            del self._requests[rid]
            defer.timeout(d)

        from twisted.internet import reactor
        timeout = reactor.callLater(
            timeout_value or REQUEST_TIMEOUT, on_timeout, None)
        return rid, ([], d, timeout, )

    def _send_request(self, line, timeout_value=None):
        d = defer.Deferred()
        rid, request = self._make_request(d, timeout_value)
        self._requests[rid] = request

        wrapper = "rid|{0}".format(rid)
        self.sendLine(wrapper)
        self.sendLine(line)
        self.sendLine(wrapper)
        return d

    def lineReceived(self, line):
        if line.startswith("<consoleN><"):
            return
        self.got_line(line.replace("\\n", '').replace("\\u0020", ' '))

    def got_line(self, line):
        if line.startswith("Command not found: rid"):
            self._process_responce_id(line)
        elif self._responce_id:
            self._requests[self._responce_id][0].append(line)
        elif self._parser:
            self._parser.parse_line(line)

    def _process_responce_id(self, line):
        try:
            rid = int(line.split('|')[1])
        except IndexError:
            log.err("RID format is malformed in \"{0}\"".format(line))
        except ValueError:
            log.err("Could not get RID value from \"{0}\"".format(line))
        else:
            if rid not in self._requests:
                log.err("Unexpected RID: {0}".format(rid))
                return
            if self._responce_id is None:
                # start request processing
                self._responce_id = rid
            elif self._responce_id == rid:
                # end request processing, send results to callback
                self._responce_id = None
                results, d, timeout = self._requests[rid]
                timeout.cancel()
                del self._requests[rid]
                d.callback(results)

    def server_info(self):
        d = self._send_request(REQ_SERVER_INFO)
        if self._parser:
            d.addCallback(self._parser.server_info)
        return d

    def mission_status(self):
        d = self._send_request(REQ_MISSION_STATUS)
        if self._parser:
            d.addCallback(self._parser.mission_status)
        return d

    def mission_load(self, mission):
        d = self._send_request(
            REQ_MISSION_LOAD.format(mission), REQUEST_MISSION_LOAD_TIMEOUT)
        if self._parser:
            d.addCallback(self._parser.mission_load)
        return d

    def mission_begin(self):
        d = self._send_request(REQ_MISSION_BEGIN)
        if self._parser:
            d.addCallback(self._parser.mission_begin)
        return d

    def mission_end(self):
        d = self._send_request(REQ_MISSION_END)
        d.addCallback(lambda _: self.mission_status())
        if self._parser:
            d.addCallback(self._parser.mission_end)
        return d

    def mission_destroy(self):
        d = self._send_request(REQ_MISSION_DESTROY)
        d.addCallback(lambda _: self.mission_status())
        if self._parser:
            d.addCallback(self._parser.mission_destroy)
        return d


class ConsoleClientFactory(ClientFactory):

    protocol = ConsoleClient

    def __init__(self, parser=None):
        self._client = None
        self.on_connecting = defer.Deferred()
        self.on_connection_lost = defer.Deferred()

    def buildProtocol(self, addr):
        self._client = ClientFactory.buildProtocol(self, addr)
        return self._client

    def clientConnectionMade(self, client):
        if self.on_connecting is not None:
            d, self.on_connecting = self.on_connecting, None
            d.callback(client)

    def clientConnectionFailed(self, connector, reason):
        if self.on_connecting is not None:
            d, self.on_connecting = self.on_connecting, None
            d.errback(reason)

    def clientConnectionLost(self, connector, reason):
        if self.on_connection_lost is not None:
            d, self.on_connection_lost = self.on_connection_lost, None
            if isinstance(reason.value, ConnectionDone):
                d.callback(None)
            else:
                d.errback(reason)


class DeviceLinkProtocol(DatagramProtocol):

    def __init__(self, address=None, parser=None):
        self.address = address
        self.parser = parser

    def datagramReceived(self, data, (host, port)):
        if data.startswith(DEVICE_LINK_PREFIXES['answer']):
            processor = self.answers_received
        elif data.startswith(DEVICE_LINK_PREFIXES['request']):
            processor = self.requests_received
        else:
            log.err("Malformed data from {0}: \"{1}\"")
            return
        prepared_data = data[1:].strip(DL_CMD_SEP)
        payloads = self._parse(prepared_data)
        processor(payloads, (host, port))

    def _parse(self, data):
        results = []
        for chunk in data.split(DL_CMD_SEP):
            command = chunk.split(DL_ARGS_SEP)
            arg = command[1:]
            results.append((command[0], arg[0] if arg else None, ))
        return results

    def answers_received(self, answers, address):
        raise NotImplementedError

    def requests_received(self, requests, address):
        raise NotImplementedError

    def send_request(self, request, address=None):
        self.send_requests([request, ], address)

    def send_requests(self, requests, address=None):
        self._send_payloads(DEVICE_LINK_PREFIXES['request'], requests, address)

    def send_answer(self, answer, address=None):
        self.send_answers([answer, ], address)

    def send_answers(self, answers, address=None):
        self._send_payloads(DEVICE_LINK_PREFIXES['answer'], answers, address)

    def _send_payloads(self, prefix, payloads, address=None):
        data = DL_CMD_SEP.join(
            [prefix, self._format(payloads), ])
        self.transport.write(data, address or self.address)

    def _format(self, payloads):
        chunks = []
        for payload in payloads:
            cmd, arg = payload
            chunks.append(
                DL_ARGS_SEP.join([cmd, str(arg)]) if arg is not None else cmd)
        return DL_CMD_SEP.join(chunks)


class DeviceLinkClient(DeviceLinkProtocol):

    cmd_group_max_size = DEVICE_LINK_CMD_GROUP_MAX_SIZE

    def __init__(self, address, parser=None):
        DeviceLinkProtocol.__init__(self, address, parser)
        # [ (opcode, deferred, timeout), ]
        self._requests = []

    def answers_received(self, answers, address):
        if address == self.address:
            for answer in answers:
                self._answer_received(answer)

    def _answer_received(self, answer):
        cmd, arg = answer
        for request in self._requests:
            opcode, d, timeout = request
            if opcode == cmd:
                timeout.cancel()
                self._requests.remove(request)
                d.callback(arg)
                break

    def _make_request(self, command, d, timeout_value=None):

        def on_timeout(_):
            self._requests.remove(request)
            defer.timeout(d)

        from twisted.internet import reactor
        timeout = reactor.callLater(
            timeout_value or REQUEST_TIMEOUT, on_timeout, None)
        request = (command[0], d, timeout, )
        return request

    def _deferred_request(self, command, timeout_value=None):
        d = defer.Deferred()
        self._requests.append(
            self._make_request(command, d, timeout_value))
        self.send_request(command)
        return d

    def _deferred_requests_iterator(self, commands):
        step = self.cmd_group_max_size
        count = len(commands)
        for i in xrange((count/step)+1):
            start = i*step
            yield commands[start:start+min((count-start), step)]

    def _deferred_requests_group(self, iterator, timeout_value=None):
        try:
            group = iterator.next()
        except StopIteration:
            return defer.succeed(None)
        else:
            dlist = []
            for cmd in group:
                d = defer.Deferred()
                self._requests.append(
                    self._make_request(cmd, d, timeout_value))
                dlist.append(d)
            self.send_requests(group)
            return defer.gatherResults(dlist)

    def _deferred_requests(self, commands, timeout_value=None):
        """
        Send requests in portions. Each next portion is send after result
        of the previous one was received. All results are collected in one
        place and are returned if there no requests has left.
        """
        all_results = []
        iterator = self._deferred_requests_iterator(commands)

        def on_results(results):
            if results:
                all_results.extend(results)
                return do_next()
            else:
                return all_results

        def do_next():
            return self._deferred_requests_group(
                iterator, timeout_value).addCallback(on_results)

        return do_next()

    def refresh_radar(self):
        self.send_request(DL_OPCODE.RADAR_REFRESH.make_command())
        return defer.succeed(None)

    def pilot_count(self):
        d = self._deferred_request(DL_OPCODE.PILOT_COUNT.make_command())
        if self.parser is not None:
            d.addCallback(self.parser.pilot_count)
        return d

    def pilot_pos(self, index):
        d = self._deferred_request(DL_OPCODE.PILOT_POS.make_command(index))
        if self.parser is not None:
            d.addCallback(self.parser.pilot_pos)
        return d

    def all_pilots_pos(self):
        d = self._all_pos(self.pilot_count, DL_OPCODE.PILOT_POS)
        if self.parser is not None:
            d.addCallback(self.parser.all_pilots_pos)
        return d

    def static_count(self):
        d = self._deferred_request(DL_OPCODE.STATIC_COUNT.make_command())
        if self.parser is not None:
            d.addCallback(self.parser.static_count)
        return d

    def static_pos(self, index):
        d = self._deferred_request(DL_OPCODE.STATIC_POS.make_command(index))
        if self.parser is not None:
            d.addCallback(self.parser.static_pos)
        return d

    def all_static_pos(self):
        d = self._all_pos(self.static_count, DL_OPCODE.STATIC_POS)
        if self.parser is not None:
            d.addCallback(self.parser.all_static_pos)
        return d

    def _all_pos(self, get_count, pos_opcode):

        def on_count(result):
            count = int(result)
            if not count:
                return []
            return self._deferred_requests([
                pos_opcode.make_command(i) for i in xrange(count)])

        return get_count().addCallback(on_count)



# -*- coding: utf-8 -*-

from il2ds_middleware.service import PilotBaseService, ObjectsBaseService


class PilotService(PilotBaseService):

    def __init__(self):
        PilotBaseService.__init__(self)
        self.joined = []
        self.left = []
        self.occupied = []
        self.weapons = []
        self.killed = []
        self.shot_down = []
        self.army_select = []
        self.to_menu = []

    def user_join(self, info):
        self.joined.append(info)

    def user_left(self, info):
        self.left.append(info)

    def seat_occupied(self, info):
        self.occupied.append(info)

    def weapons_loaded(self, info):
        self.weapons.append(info)

    def was_killed(self, info):
        self.killed.append(info)

    def was_shot_down(self, info):
        self.shot_down.append(info)

    def selected_army(self, info):
        self.army_select.append(info)

    def went_to_menu(self, info):
        self.to_menu.append(info)


class ObjectsService(ObjectsBaseService):

    def __init__(self):
        ObjectsBaseService.__init__(self)
        self.destroyed = []

    def was_destroyed(self, info):
        self.destroyed.append(info)



# -*- coding: utf-8 -*-

import re

from twisted.python import log

from zope.interface import implementer

from il2ds_middleware.constants import MISSION_STATUS, PILOT_LEAVE_REASON
from il2ds_middleware.interface.parser import (IDeviceLinkParser,
    IEventLogParser, IConsoleParser, )
from il2ds_middleware.regex import *


@implementer(IConsoleParser)
class ConsoleParser(object):

    _buffer = None

    def __init__(self, pilot_service):
        self.pilot_service = pilot_service

    def parse_line(self, line):
        if self.user_joined(line):
            return
        elif self.user_left(line):
            return

    def server_info(self, lines):
        result = {}
        for line in lines:
            key, value = line.split(':')
            result[key.strip().lower()] = value.strip()
        return result

    def mission_status(self, lines):
        for line in lines:
            if line == "Mission NOT loaded":
                return (MISSION_STATUS.NOT_LOADED, None, )
            elif line.endswith("is Loaded"):
                return (MISSION_STATUS.LOADED, line.split()[1], )
            elif line.endswith("is Playing"):
                return (MISSION_STATUS.PLAYING, line.split()[1], )
        return lines

    mission_load = mission_destroy = mission_status
    mission_begin = mission_end = mission_status

    def user_joined(self, line):
        m = re.match(RX_USER_JOIN, line)
        if not m:
            return False
        else:
            groups = m.groups()
            info = {
                'channel': int(groups[0]),
                'ip': groups[1],
                'callsign': groups[2],
            }
            self.pilot_service.user_join(info)
            return True

    def user_left(self, line):
        m = re.match(RX_USER_LEFT, line)
        if m:
            groups = m.groups()
            reason = (PILOT_LEAVE_REASON.KICKED if 'kicked' in groups[2]
                else PILOT_LEAVE_REASON.DISCONNECTED)
            self._buffer = {
                'ip': groups[0],
                'channel': int(groups[1]),
                'reason': reason,
            }
            return True
        if line.endswith("has left the game.") and self._buffer:
            info, self._buffer = self._buffer, None
            info['callsign'] = line.split()[2]
            self.pilot_service.user_left(info)
            return True
        return False


@implementer(IEventLogParser)
class EventLogParser(object):

    def __init__(self, (pilot_service, objects_service)):
        ps = pilot_service
        obs = objects_service
        self.parsers = (
            (RX_SEAT_OCCUPIED, self.seat_occupied, ps.seat_occupied),
            (RX_WEAPONS_LOADED, self.weapons_loaded, ps.weapons_loaded),
            (RX_KILLED, self.was_killed, ps.was_killed),
            (RX_SHOT_DOWN, self.was_shot_down, ps.was_shot_down),
            (RX_SELECTED_ARMY, self.selected_army, ps.selected_army),
            (RX_WENT_TO_MENU, self.went_to_menu, ps.went_to_menu),
            (RX_DESTROYED, self.was_destroyed, obs.was_destroyed),
        )

    def parse_line(self, line):
        for rx, parser, dst in self.parsers:
            m = re.match(rx, line)
            if m:
                dst(parser(m.groups()))
                return

    def seat_occupied(self, groups):
        return {
            'callsign': groups[0],
            'aircraft': groups[1],
            'seat': int(groups[2]),
            'pos': self._get_pos((groups[3], groups[4])),
        }

    def weapons_loaded(self, groups):
        return {
            'callsign': groups[0],
            'aircraft': groups[1],
            'weapons': groups[2],
            'fuel': int(groups[3]),
        }

    def was_killed(self, groups):
        return {
            'callsign': groups[0],
            'aircraft': groups[1],
            'seat': int(groups[2]),
            'pos': self._get_pos((groups[3], groups[4])),
        }

    def was_shot_down(self, groups):
        return {
            'victim': {
                'callsign': groups[0],
                'aircraft': groups[1],
            },
            'attacker': {
                'callsign': groups[2],
                'aircraft': groups[3],
            },
            'pos': self._get_pos((groups[4], groups[5])),
        }

    def selected_army(self, groups):
        return {
            'callsign': groups[0],
            'army': groups[1],
            'pos': self._get_pos((groups[2], groups[3])),
        }

    def went_to_menu(self, groups):
        return {
            'callsign': groups[0],
        }

    def was_destroyed(self, groups):
        return {
            'victim': groups[0],
            'attacker': groups[1],
            'pos': self._get_pos((groups[2], groups[3])),
        }

    def _get_pos(self, (x, y)):
        return {
            'x': float(x),
            'y': float(y),
        }


@implementer(IDeviceLinkParser)
class DeviceLinkParser(object):

    def pilot_count(self, data):
        return int(data)

    def pilot_pos(self, data):
        return self._parse_pos(data, 'callsign')

    def all_pilots_pos(self, datas):
        return map(self.pilot_pos, datas)

    static_count = pilot_count

    def static_pos(self, data):
        return self._parse_pos(data)

    def all_static_pos(self, datas):
        return map(self.static_pos, datas)

    def _parse_pos(self, data, name_attr='name'):
        idx, info = data.split(':')
        attr, x, y, z = info.split(';')
        return {
            'idx': int(idx),
            name_attr: attr,
            'pos': {
                'x': int(x),
                'y': int(y),
                'z': int(z),
            },
        }
